@@ -1,34 +1,19 @@
 #!/usr/bin/env python3
 # -*- coding: utf-8 -*-
 """
-Created on Fri Apr 16 16:39:21 2021
+The linkage module defines useful classes for linkage definition.
+
+Created on Fri Apr 16 16:39:21 2021.
 
 @author: HugoFara
 """
+import abc
 from math import atan2, gcd, tau
 from abc import ABC
+from .exceptions import HypostaticError, UnbuildableError
 from .geometry import sqr_dist, circle_intersect, cyl_to_cart
 
 
-class UnbuildableError(Exception):
-    """Should be raised when the constraints cannont be solved."""
-
-    def __init__(self, joint, message='Unable to solve constraints'):
-        self.joint = joint
-        super().__init__(message)
-
-    def __str__(self):
-        return f"{self.joint} cannot be solved"
-
-
-class HypostaticError(Exception):
-    """Exception when the system can be solved in multiple ways."""
-
-    def __init__(self, linkage, message='The system is hypostatic!'):
-        self.linkage = linkage
-        super().__init__(message)
-
-
 class Joint(ABC):
     """
     Geometric constraint expressed by two joints.
@@ -44,20 +29,35 @@ class Joint(ABC):
 
         Arguments
         ---------
-        * x: joint axis, scalar
-        * y: joint ordinate
-        * joint0: first reference to use, usually a Joint
-        * joint1: second reference, same behavior as joint0
-        * name: unique name
+        x : float, optional
+            Position on horizontal axis. The default is 0.
+        y : float, optional
+            Position on vertical axis. The default is O.
+        name : str, optional
+            Friendly name for human readability. The default is None.
+        joint0 : Union[Joint, tuple[float]], optional
+            Linked pivot joint 1 (geometric constraints). The default is None.
+        joint1 : Union[Joint, tuple[float]], optional
+            Other pivot joint linked. The default is None.
         """
         self.x, self.y = x, y
-        self.joint0, self.joint1 = joint0, joint1
-        self.name = name or str(id(self))
+        if joint0 is None or isinstance(joint0, Joint):
+            self.joint0 = joint0
+        else:
+            self.joint0 = Static(*joint0)
+        if joint1 is None or isinstance(joint1, Joint):
+            self.joint1 = joint1
+        else:
+            self.joint1 = Static(*joint1)
+        self.name = name
+        if name is None:
+            self.name = str(id(self))
 
     def __repr__(self):
         """Represent object with class name, coordinates, name and state."""
-        return "{}(x={}, y={}, name={})".format(self.__class__.__name__,
-                                                self.x, self.y, self.name)
+        return "{}(x={}, y={}, name={})".format(
+            self.__class__.__name__, self.x, self.y, self.name
+        )
 
     def __get_joints__(self):
         """Return constraint joints as a tuple."""
@@ -74,6 +74,20 @@ class Joint(ABC):
         else:
             self.x, self.y = args[0], args[1]
 
+    @abc.abstractmethod
+    def get_constraints(self):
+        """Return geometric constraints applying to this Joint."""
+        raise NotImplementedError(
+            "You can't call a constraint from an abstract class."
+        )
+
+    @abc.abstractmethod
+    def set_constraints(self, *args):
+        """Set geometric constraints applying to this Joint."""
+        raise NotImplementedError(
+            "You can't set a constraint from an abstract class."
+        )
+
 
 class Static(Joint):
     """
@@ -85,20 +99,38 @@ class Static(Joint):
     __slots__ = tuple()
 
     def __init__(self, x=0, y=0, name=None):
+        """
+        A Static joint is a point in space to use as anchor by other joints.
+
+        It is NOT a kind of joint as viewed in engineering terms!
+
+        x : float, optional
+            Position on horizontal axis. The default is 0.
+        y : float, optional
+            Position on vertical axis. The default is O.
+        name : str, optional
+            Friendly name for human readability. The default is None.
+        """
         super().__init__(x, y, name=name)
 
     def reload(self):
-        """Do nothing, for consistence only."""
-        return
+        """Do nothing, for consistency only."""
+        pass
+
+    def get_constraints(self):
+        """Return an empty tuple."""
+        return tuple()
 
     def set_constraints(self, *args):
-        """Do nothing, for consistence only."""
-        return
+        """Do nothing, for consistency only."""
+        pass
 
     def set_anchor0(self, joint):
+        """First joint anchor."""
         self.joint0 = joint
 
     def set_anchor1(self, joint):
+        """Second joint anchor."""
         self.joint1 = joint
 
 
@@ -114,11 +146,22 @@ class Fixed(Joint):
 
         Arguments
         ---------
-        * joint0: first reference joint,
-        * joint1: second reference joint.
-        * distance: to keep constant between joint0 and self.
-        * angle: It is the angle (joint1, joint0, self).
-        Should be in radian and in trigonometric order.
+        x : float, optional
+            Position on horizontal axis. The default is 0.
+        y : float, optional
+            Position on vertical axis. The default is O.
+        name : str, optional
+            Friendly name for human readability. The default is None.
+        joint0 : Union[Joint, tuple[float]], optional
+            Linked pivot joint 1 (geometric constraints). The default is None.
+        joint1 : Union[Joint, tuple[float]], optional
+            Other pivot joint linked. The default is None.
+        distance : float, optional
+            Distance to keep constant between joint0 and self. The default is
+            None.
+        angle : float, optional
+         Angle (joint1, joint0, self). Should be in radian and in trigonometric
+         order. The default is None.
         """
         super().__init__(x, y, joint0, joint1, name)
         self.angle = angle
@@ -144,15 +187,21 @@ class Fixed(Joint):
         self.x, self.y = cyl_to_cart(self.r, self.angle + rot,
                                      self.joint0.coord())
 
+    def get_constraints(self):
+        """Return the constraining distance and angle parameters."""
+        return self.r, self.angle
+
     def set_constraints(self, distance=None, angle=None):
         """Set geometric constraints."""
         self.r, self.angle = distance or self.r, angle or self.angle
 
     def set_anchor0(self, joint, distance=None, angle=None):
+        """First joint anchor and characterisitcs."""
         self.joint0 = joint
         self.set_constraints(distance, angle)
 
     def set_anchor1(self, joint):
+        """Second joint anchor."""
         self.joint1 = joint
 
 
@@ -166,16 +215,22 @@ class Pivot(Joint):
         """
         Set point position, parents, and if it is fixed for this turn.
 
-        Arguments:
+        Arguments
         ---------
-        * x: position on horizontal axis
-        * y: position on vertical axis
-        * name: friendly name for human readability
-        * joint0: linked pivot joint 1 (geometric constraints). A Joint
-        representing the center of a pivot joint.
-        * joint1: other pivot joint linked.
-        * distance0: distance from joint0 to the current Joint
-        * distance1: distance from joint1 to the current Joint.
+        x : float, optional
+            Position on horizontal axis. The default is 0.
+        y : float, optional
+            Position on vertical axis. The default is O.
+        name : str, optional
+            Friendly name for human readability. The default is None.
+        joint0 : Union[Joint, tuple[float]], optional
+            Linked pivot joint 1 (geometric constraints). The default is None.
+        joint1 : Union[Joint, tuple[float]], optional
+            Other pivot joint linked. The default is None.
+        distance0 : float, optional
+            Distance from joint0 to the current Joint. The default is None.
+        distance1 : float, optional
+            Distance from joint1 to the current Joint. The default is None.
         """
         super().__init__(x, y, joint0, joint1, name)
         self.r0, self.r1 = distance0, distance1
@@ -195,9 +250,9 @@ class Pivot(Joint):
         Circle is a tuple (abscisse, ordinate, radius).
         """
         if self.joint0 is joint:
-            return joint.x, joint.y, self.ref1[1]
-        elif self.joint1 is joint:
-            return joint.x, joint.y, self.ref2[1]
+            return joint.x, joint.y, self.r0
+        if self.joint1 is joint:
+            return joint.x, joint.y, self.r1
         raise ValueError(f'{joint} is not in joints of {self}')
 
     def reload(self):
@@ -235,16 +290,50 @@ class Pivot(Joint):
                                                        self.coord()))
                 self.x, self.y = [i + j * coco[1][2] for i, j in zip(coco[1],
                                   vect)]
+    def get_constraints(self):
+        """Return the two constraining distances of this joint."""
+        return self.r0, self.r1
 
     def set_constraints(self, distance0=None, distance1=None):
         """Set geometric constraints."""
         self.r0, self.r1 = distance0 or self.r0, distance1 or self.r1
 
     def set_anchor0(self, joint, distance=None):
+        """
+        Set the first anchor for this Joint.
+
+        Parameters
+        ----------
+        joint : Union[Joint, tuple[float]]
+            The joint to use as achor.
+        distance : float, optional
+            Distance to keep constant from the anchor. The default is None.
+
+        Returns
+        -------
+        None.
+
+        """
         self.joint0 = joint
         self.set_constraints(distance0=distance)
 
     def set_anchor1(self, joint, distance=None):
+        """
+        Set the second anchor for this Joint.
+
+        Parameters
+        ----------
+        joint : Union[Joint, tuple[float]]
+            The joint to use as achor.
+        distance : float, optional
+            Distance to keep constant from the anchor. The default is None.
+
+        Returns
+        -------
+        None.
+
+        """
+        self.joint1 = joint
         self.set_constraints(distance1=distance)
 
 
@@ -265,7 +354,7 @@ class Crank(Joint):
             The default is None.
         y : float, optional
             initial vertical position. The default is None.
-        joint0 : Joint, optional
+        joint0 : Union[Joint, tuple[float]], optional
             first reference joint. The default is None.
         distance : float, optional
             distance to keep between joint0 and self. The default is None.
@@ -296,11 +385,16 @@ class Crank(Joint):
         self.x, self.y = cyl_to_cart(self.r, rot + self.angle * dt,
                                      self.joint0.coord())
 
+    def get_constraints(self):
+        """Return the distance to the center of rotation."""
+        return (self.r,)
+
     def set_constraints(self, distance=None, *args):
         """Set geometric constraints, only self.r is affected."""
         self.r = distance or self.r
 
     def set_anchor0(self, joint, distance=None):
+        """First joint anchor and fixed distance."""
         self.joint0 = joint
         self.set_constraints(distance=distance)
 
@@ -321,12 +415,19 @@ class Linkage():
 
         Arguments
         ---------
-        * joints: all Joint to be part of the linkage
-        * order: sequence that manually define resolution order for each step.
-        Should be elements of joints if provided.
-        * name: linkage name
+        joints : list[Joint]
+            All Joint to be part of the linkage
+        order : list[Joint]
+            Sequence to manually define resolution order for each step.
+            Should be a subset of joints.
+            Automatic computed order is experimental! The default is None.
+        name : str, optional
+            Human-readable name for the Linkage. If None, take the value
+            str(id(self)). The default is None.
         """
-        self.name = str(name or id(self))
+        self.name = name
+        if name is None:
+            self.name = str(id(self))
         self.joints = tuple(joints)
         self._cranks = tuple(j for j in joints if isinstance(j, Crank))
         if order:
@@ -337,7 +438,7 @@ class Linkage():
         self._solve_order = order
 
     def __find_solving_order__(self):
-        """Automatically finds solving order."""
+        """Find solving order automatically (experimental)."""
         # TODO : test it
         solvable = [j for j in self.joints if isinstance(j, Static)]
         # True of new joints where added in the current pass
@@ -358,7 +459,7 @@ class Linkage():
             raise HypostaticError(
                 'Unable to determine automatic order!'
                 'Those joints are left unsolved:'
-                ','.join(j for j in self.joints if j not in solvable)
+                ','.join(str(j) for j in self.joints if j not in solvable)
                 )
         self._solve_order = tuple(solvable)
         raise NotImplementedError('Unable to determine automatic order')
@@ -368,10 +469,13 @@ class Linkage():
         """
         Redifine linkage joints and given intial positions to joints.
 
-        pos: a tuple of initial positions, in the same order as linka["order"],
-        for linka["crank"] at position (0, -1). Fixed_Joint WON'T be modified.
-        Coordinates does not need to be precise, they will allow us the best
-        fitting position between all possible positions satifying constraints.
+        Parameters
+        ----------
+        pos : tuple[tuple[int]]
+            Initial positions for each joint in self.joints.
+            Coordinates does not need to be precise, they will allow us the best
+            fitting position between all possible positions satifying
+            constraints.
         """
         if not hasattr(self, '_solve_order'):
             self.__find_solving_order__()
@@ -380,16 +484,41 @@ class Linkage():
         # Parents joint do not have children.
         if pos is not None:
             # Defintition of initial coordinates
-            for p, j in zip(pos, self.joints):
-                j.set_coord(p)
+            self.set_coords(pos)
 
-    def get_pos(self):
+    def get_coords(self):
         """Return positions of all elements of the system."""
-        return (j.coord() for j in self.joints)
+        return [j.coord() for j in self.joints]
+
+    def set_coords(self, coords):
+        """Set coordinatess for all joints of the linkage."""
+        for joint, coord in zip(self.joints, coords):
+            joint.set_coord(coord)
 
     def hyperstaticity(self):
         """Return the hyperstaticity degree of the linkage in 2D."""
-        raise NotImplementedError('Not added in this version!')
+        # TODO : test it
+        # We have at least the frame
+        solids = 1
+        mobilities = 1
+        kinematic_indetermined = 0
+        for j in self.joints:
+            if isinstance(j, (Static, Fixed)):
+                pass
+            elif isinstance(j, Crank):
+                solids += 1
+                kinematic_indetermined += 2
+            elif isinstance(j, Pivot):
+                solids += 1
+                # A Pivot Joint create at least two pivots
+                kinematic_indetermined += 4
+                if not hasattr(j, 'joint1') or j.joint1 is None:
+                    mobilities += 1
+                else:
+                    solids += 1
+                    kinematic_indetermined += 2
+
+        return 3 * (solids - 1) - kinematic_indetermined + mobilities
 
     def step(self, iterations=None, dt=1):
         """
@@ -413,7 +542,7 @@ class Linkage():
         """
         if iterations is None:
             iterations = self.get_rotation_period()
-        for i in range(iterations):
+        for _ in range(iterations):
             for j in self._solve_order:
                 if isinstance(j, Crank):
                     j.reload(dt)
@@ -421,25 +550,59 @@ class Linkage():
                     j.reload()
             yield tuple(j.coord() for j in self.joints)
 
-    def set_coords(self, coords):
-        """Set coordinatess for all joints of the linkage."""
-        for j, c in zip(self.joints, coords):
-            j.set_coord(c)
+    def get_num_constraints(self, flat=True):
+        """
+        Return numeric constraints of this linkage.
 
-    def set_num_constraints(self, constraints):
+        Parameters
+        ----------
+        flat : bool
+            Whether to force one-dimensionnal representation of constraints.
+            The default is True.
+
+        Returns
+        -------
+        constraints : list
+            List of geometric constraints.
+        """
+        constraints = []
+        for joint in self.joints:
+            if flat:
+                constraints.extend(joint.get_constraints())
+            else:
+                constraints.append(joint.get_constraints())
+        return constraints
+
+    def set_num_constraints(self, constraints, flat=True):
         """
         Set numeric constraints for this linkage.
 
         Numeric constraints are distances or angles between joints.
 
-        Arguments
-        ---------
-        * constraints: a sequence of tuples of digits. Should be in same order
-        as self.joints. Each element will be passed to the set_constraints
-        method of each correspondig Joint.
+        Parameters
+        ----------
+        constraints : sequence
+            Sequence of constraints to pass to the joints.
+        flat : bool
+            If True, should be a one-dimensionnal sequence of floats.
+            If False, should be a sequence of tuples of digits. Each element
+            will be passed to the set_constraints method of each correspondig
+            Joint.
+            The default is True.
         """
-        for j, c in zip(self.joints, constraints):
-            j.set_constraints(*c)
+        if flat:
+            # Is in charge of redistributing constraints
+            dispatcher = iter(constraints)
+            for joint in self.joints:
+                if isinstance(joint, Static):
+                    pass
+                elif isinstance(joint, Crank):
+                    joint.set_constraints(next(dispatcher))
+                elif isinstance(joint, (Fixed, Pivot)):
+                    joint.set_constraints(next(dispatcher), next(dispatcher))
+        else:
+            for joint, constraint in zip(self.joints, constraints):
+                joint.set_constraints(*constraint)
 
     def get_rotation_period(self):
         """
